'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var _rollupPluginBabelHelpers = require('./_virtual/_rollupPluginBabelHelpers.js');
var events = require('events');
var path = require('path');
var graphql = require('graphql');
var chalk = require('chalk');
var fsExtra = require('fs-extra');
var graphqlConfig = require('graphql-config');
var graphqlConfigUtilities = require('graphql-config-utilities');
var graphqlToolUtilities = require('graphql-tool-utilities');
var types = require('./types.js');
var index = require('./print/schema/index.js');
var index$1 = require('./print/document/index.js');
var graphqlFilesystem = require('./filesystem/graphql-filesystem.js');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var chalk__default = /*#__PURE__*/_interopDefaultLegacy(chalk);

const _excluded = ["cwd"];
class Builder extends events.EventEmitter {
  // workspace graphql configuration
  // see: https://github.com/prisma/graphql-config
  // projectName -> {filePath -> document}
  // NOTE: projectName can be undefined for nameless graphql-config projects
  constructor(_ref) {
    var _this;

    let {
      cwd
    } = _ref,
        options = _rollupPluginBabelHelpers.objectWithoutProperties(_ref, _excluded);

    super();
    _this = this;
    this.options = void 0;
    this.config = void 0;
    this.documentMapByProject = new Map();
    this.filesystem = void 0;
    this.schemaCache = new Map();
    this.activeDocumentTypesJobId = 0;

    this.handleDocumentUpdate = /*#__PURE__*/function () {
      var _ref2 = _rollupPluginBabelHelpers.asyncToGenerator(function* (filePath, projectConfig) {
        try {
          yield _this.updateDocumentForFile(filePath, projectConfig);
        } catch (error) {
          _this.emit('error', error);

          return;
        }

        yield _this.generateDocumentTypes();
      });

      return function (_x, _x2) {
        return _ref2.apply(this, arguments);
      };
    }();

    this.handleDocumentDelete = /*#__PURE__*/function () {
      var _ref3 = _rollupPluginBabelHelpers.asyncToGenerator(function* (filePath, projectConfig) {
        const documents = _this.documentMapByProject.get(projectConfig.name);

        if (documents) {
          documents.delete(filePath);
        }

        yield _this.generateDocumentTypes();
      });

      return function (_x3, _x4) {
        return _ref3.apply(this, arguments);
      };
    }();

    this.handleSchemaUpdate = /*#__PURE__*/function () {
      var _ref4 = _rollupPluginBabelHelpers.asyncToGenerator(function* (input) {
        const schemaPaths = typeof input === 'string' ? [input] : input;

        try {
          _this.emit('start:schema');

          yield Promise.all(schemaPaths.map(schemaPath => _this.generateSchemaTypes(schemaPath)));

          _this.emit('end:schema');

          yield _this.generateDocumentTypes();
        } catch (error) {// intentional noop
        }
      });

      return function (_x5) {
        return _ref4.apply(this, arguments);
      };
    }();

    this.options = options;
    const configPath = cwd ? path.resolve(cwd) : undefined;
    this.config = options.config ? options.config : graphqlConfig.loadConfigSync({
      rootDir: configPath
    });
  }

  once(event, handler) {
    return super.once(event, handler);
  }

  on(event, handler) {
    return super.on(event, handler);
  }

  emit(event, ...args) {
    return super.emit(event, ...args);
  }

  run({
    watch: watchGlobs = false,
    graphQLFilesystem
  } = {}) {
    var _this2 = this;

    return _rollupPluginBabelHelpers.asyncToGenerator(function* () {
      let schemaPaths;

      try {
        schemaPaths = graphqlConfigUtilities.getGraphQLSchemaPaths(_this2.config);
      } catch (error) {
        _this2.emit('error', error);

        return;
      }

      const filesystem = graphQLFilesystem !== null && graphQLFilesystem !== void 0 ? graphQLFilesystem : yield defaultFilesytem();
      _this2.filesystem = filesystem;

      if (watchGlobs) {
        filesystem.on('change:document', _this2.handleDocumentUpdate).on('delete:document', _this2.handleDocumentDelete).on('change:schema', _this2.handleSchemaUpdate); // wait for all watchers to be ready

        yield filesystem.watch(_this2.config);
      }

      try {
        _this2.emit('start:schema');

        yield Promise.all(schemaPaths.map(schemaPath => _this2.generateSchemaTypes(schemaPath)));

        _this2.emit('end:schema');
      } catch (error) {
        _this2.emit('error', error);

        return;
      }

      try {
        yield Promise.all(graphqlConfigUtilities.getGraphQLProjects(_this2.config).map(projectConfig => _this2.updateDocumentsForProject(filesystem, projectConfig)));
      } catch (error) {
        _this2.emit('error', error);

        return;
      }

      yield _this2.generateDocumentTypes();
    })();
  }

  stop() {
    if (this.filesystem && typeof this.filesystem.dispose === 'function') {
      this.filesystem.dispose();
    }
  }

  generateSchemaTypes(schemaPath) {
    var _this3 = this;

    return _rollupPluginBabelHelpers.asyncToGenerator(function* () {
      const projectConfig = graphqlConfigUtilities.getGraphQLProjectForSchemaPath(_this3.config, schemaPath);
      const schemaTypesPath = getSchemaTypesPath(projectConfig, _this3.options);
      const definitions = index.generateSchemaTypes(yield _this3.getSchema(projectConfig), _this3.options);
      yield fsExtra.mkdirp(schemaTypesPath);
      yield Promise.all(Array.from(definitions.entries()).map(([fileName, definition]) => fsExtra.writeFile(path.join(schemaTypesPath, fileName), definition)));

      _this3.emit('build:schema', {
        schemaPath,
        schemaTypesPath
      });
    })();
  }

  isDocumentTypesJobCancelled(jobId) {
    return this.activeDocumentTypesJobId !== jobId;
  }

  generateDocumentTypes() {
    var _this4 = this;

    return _rollupPluginBabelHelpers.asyncToGenerator(function* () {
      // Document generation can be called often via:
      // – Projects with small + large schemas calling `handleSchemaUpdate`
      //   during schema refreshes
      // - Projects with many .graphql files triggering multiple
      //  `handleDocumentUpdate` calls during pulls (or rebases)
      //
      // The above scenarios do not provide a predictable debounce window that
      // would prevent multiple document generation jobs running, so generation
      // is allowed to immediately proceed, but each "job" checks itself often to
      // see if it should stop.
      const jobId = ++_this4.activeDocumentTypesJobId;
      return new Promise((resolve, reject) => {
        _this4.once('end:docs', () => {
          resolve();
        });

        _this4.once('error', () => reject());

        _this4._generateDocumentTypes(jobId).catch(err => reject(err));
      });
    })();
  }

  _generateDocumentTypes(jobId) {
    var _this5 = this;

    return _rollupPluginBabelHelpers.asyncToGenerator(function* () {
      _this5.emit('start:docs');

      _this5.checkForDuplicateOperations();

      _this5.checkForDuplicateFragments();

      yield Promise.all(Array.from(_this5.documentMapByProject.entries()).map(([projectName, documents]) => _this5.generateDocumentTypesForProject(_this5.config.getProject(projectName), documents, jobId)));
      if (_this5.isDocumentTypesJobCancelled(jobId)) return;

      _this5.emit('end:docs');
    })();
  }

  checkForDuplicateOperations() {
    getDuplicateOperations(this.documentMapByProject).forEach(({
      projectName,
      duplicates
    }) => {
      if (duplicates.length) {
        duplicates.forEach(({
          operationName,
          filePaths
        }) => {
          const message = `GraphQL operations must have a unique name. The operation ${chalk__default["default"].bold(operationName)} is declared in:\n ${filePaths.sort().join('\n ')}${projectName ? ` (${chalk__default["default"].bold(projectName)})` : ''}`;
          this.emit('error', new Error(message));
        });
      }
    });
  }

  checkForDuplicateFragments() {
    getDuplicateFragments(this.documentMapByProject).forEach(({
      projectName,
      duplicates
    }) => {
      if (duplicates.length) {
        duplicates.forEach(({
          fragmentName,
          filePaths
        }) => {
          const message = `GraphQL fragments must have a unique name. The fragment ${chalk__default["default"].bold(fragmentName)} is declared in:\n ${filePaths.sort().join('\n ')}${projectName ? ` (${chalk__default["default"].bold(projectName)})` : ''}`;
          this.emit('error', new Error(message));
        });
      }
    });
  }

  generateDocumentTypesForProject(projectConfig, documents, jobId) {
    var _this6 = this;

    return _rollupPluginBabelHelpers.asyncToGenerator(function* () {
      let ast;
      if (_this6.isDocumentTypesJobCancelled(jobId)) return;

      try {
        ast = graphqlToolUtilities.compile(yield _this6.getSchema(projectConfig), graphql.concatAST(Array.from(documents.values())));
      } catch (error) {
        _this6.emit('error', error);

        return;
      }

      if (_this6.isDocumentTypesJobCancelled(jobId)) return;

      try {
        for (const file of groupOperationsAndFragmentsByFile(ast).values()) {
          if (_this6.isDocumentTypesJobCancelled(jobId)) return;
          yield _this6.writeDocumentFile(file, ast, projectConfig);
        }
      } catch (error) {// intentional noop
      }
    })();
  }

  writeDocumentFile(file, ast, projectConfig) {
    var _this7 = this;

    return _rollupPluginBabelHelpers.asyncToGenerator(function* () {
      const definitionPath = `${file.path}.d.ts`;
      yield fsExtra.writeFile(definitionPath, _this7.getDocumentDefinition(file, ast, projectConfig));

      _this7.emit('build:docs', {
        documentPath: file.path,
        definitionPath,
        operation: file.operation,
        fragments: file.fragments
      });
    })();
  }

  getDocumentDefinition(file, ast, projectConfig) {
    try {
      return index$1.printDocument(file, ast, {
        enumFormat: this.options.enumFormat,
        exportFormat: this.options.exportFormat,
        addTypename: this.options.addTypename,
        schemaTypesPath: getSchemaTypesPath(projectConfig, this.options)
      });
    } catch ({
      message
    }) {
      const error = new Error(`Error in ${file.path}: ${message[0].toLowerCase()}${message.slice(1)}`);
      this.emit('error', error);
      throw error;
    }
  }

  updateDocumentsForProject(filesystem, projectConfig) {
    var _this8 = this;

    return _rollupPluginBabelHelpers.asyncToGenerator(function* () {
      const filePaths = yield filesystem.getGraphQLProjectIncludedFilePaths(projectConfig);
      return Promise.all(filePaths.map(filePath => _this8.updateDocumentForFile(filePath, projectConfig)));
    })();
  }

  updateDocumentForFile(filePath, projectConfig) {
    var _this9 = this;

    return _rollupPluginBabelHelpers.asyncToGenerator(function* () {
      const contents = yield fsExtra.readFile(filePath, 'utf8');
      return _this9.setDocumentForFilePath(filePath, projectConfig, contents);
    })();
  }

  setDocumentForFilePath(filePath, projectConfig, contents) {
    let documents = this.documentMapByProject.get(projectConfig.name);

    if (!documents) {
      documents = new Map();
      this.documentMapByProject.set(projectConfig.name, documents);
    }

    if (contents.trim().length === 0) {
      return undefined;
    }

    const document = graphql.parse(new graphql.Source(contents, filePath));
    documents.set(filePath, document);
    return document;
  }

  getSchema(projectConfig) {
    const {
      name
    } = projectConfig;

    if (!this.schemaCache.has(name)) {
      const schemaPromise = projectConfig.getSchema();
      this.schemaCache.set(name, schemaPromise);
    }

    return this.schemaCache.get(name);
  }

}

function getSchemaTypesPath(projectConfig, options) {
  if (typeof projectConfig.extensions.schemaTypesPath === 'string') {
    return graphqlConfigUtilities.resolvePathRelativeToConfig(projectConfig, projectConfig.extensions.schemaTypesPath);
  }

  return graphqlConfigUtilities.resolvePathRelativeToConfig(projectConfig, path.join(options.schemaTypesPath, `${projectConfig.name ? `${projectConfig.name}-` : ''}types`));
}

function defaultFilesytem() {
  return _defaultFilesytem.apply(this, arguments);
}

function _defaultFilesytem() {
  _defaultFilesytem = _rollupPluginBabelHelpers.asyncToGenerator(function* () {
    const {
      DefaultGraphQLFilesystem
    } = yield Promise.resolve().then(function () { return require('./filesystem/default-graphql-filesystem.js'); });
    return new DefaultGraphQLFilesystem();
  });
  return _defaultFilesytem.apply(this, arguments);
}

function groupOperationsAndFragmentsByFile({
  operations,
  fragments
}) {
  return Object.values(operations).concat(Object.values(fragments)).reduce((map, item) => {
    if (!item.filePath) {
      return map;
    }

    let file = map.get(item.filePath);

    if (!file) {
      file = {
        path: item.filePath,
        operation: undefined,
        fragments: []
      };
      map.set(item.filePath, file);
    }

    if (graphqlToolUtilities.isOperation(item)) {
      file.operation = item;
    } else {
      file.fragments.push(item);
    }

    return map;
  }, new Map());
}

function getDuplicateOperations(documentsMapByProject) {
  return Array.from(documentsMapByProject.entries()).map(([projectName, documents]) => {
    return {
      projectName,
      duplicates: getDuplicateProjectOperations(documents)
    };
  });
}

function getDuplicateFragments(documentsMapByProject) {
  return Array.from(documentsMapByProject.entries()).map(([projectName, documents]) => {
    return {
      projectName,
      duplicates: getDuplicateProjectFragments(documents)
    };
  });
}

function getDuplicateProjectOperations(documents) {
  const operations = new Map();
  Array.from(documents.entries()).forEach(([filePath, document]) => {
    document.definitions.filter(isOperationDefinition).forEach(definition => {
      const {
        name
      } = definition;

      if (name && name.value) {
        const map = operations.get(name.value);

        if (map) {
          map.add(filePath);
        } else {
          operations.set(name.value, new Set([filePath]));
        }
      }
    });
  });
  return Array.from(operations.entries()).filter(([, filePaths]) => filePaths.size > 1).map(([operationName, filePath]) => {
    return {
      operationName,
      filePaths: Array.from(filePath)
    };
  });
}

function getDuplicateProjectFragments(documents) {
  const fragments = new Map();
  Array.from(documents.entries()).forEach(([filePath, document]) => {
    document.definitions.filter(isFragmentDefinition).forEach(definition => {
      const {
        name
      } = definition;

      if (name && name.value) {
        const map = fragments.get(name.value);

        if (map) {
          map.add(filePath);
        } else {
          fragments.set(name.value, new Set([filePath]));
        }
      }
    });
  });
  return Array.from(fragments.entries()).filter(([, filePaths]) => filePaths.size > 1).map(([fragmentName, filePath]) => {
    return {
      fragmentName,
      filePaths: Array.from(filePath)
    };
  });
}

function isOperationDefinition(definition) {
  return definition.kind === 'OperationDefinition';
}

function isFragmentDefinition(definition) {
  return definition.kind === 'FragmentDefinition';
}

Object.defineProperty(exports, 'EnumFormat', {
  enumerable: true,
  get: function () { return types.EnumFormat; }
});
Object.defineProperty(exports, 'ExportFormat', {
  enumerable: true,
  get: function () { return types.ExportFormat; }
});
exports.AbstractGraphQLFilesystem = graphqlFilesystem.AbstractGraphQLFilesystem;
exports.Builder = Builder;
